/*
 * slink 0.01, Tool for replacement of duplicated files with symlinks.
 *
 * Copyright (C) 2005 Jindrich Novy (jnovy@redhat.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <inttypes.h>
#include <limits.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stddef.h>
#include <unistd.h>
#include <dirent.h>
#include <getopt.h>
#include <string.h>
#include <ctype.h>
#include <utime.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <errno.h>

#include <zlib.h>

#include "md5.h"

/* ALLOC*sizeof(slink_base) blocks are (re)allocated in single request whenever needed. */
#define ALLOC 0x1000

/* Buffer for reading. */
#define R_BUF 0x1000

/* Number of directory entries allocated in single request. */
#define R_DIR 0x100

/* Temp suffix */
#define TEMP_SUFFIX ".XXXXXX"

/* Maximal length in bytes of a single line in repository. */
#define MAX_LINE_SIZE 0x1000

/* Maximal filename size with full path */
#define MAX_FILENAME_SIZE 0x200

enum {
	SLINK_ALIGNED,
	SLINK_UNALIGNED
};

typedef struct {
	char *file;
	char *purename;
	char *package;
	char *md5;
	struct stat *st;
} slink_base;

slink_base *base = NULL;
uint32_t slink_files = 0;
uint32_t slink_maxfiles = 0;
off_t slink_totsize = 0;
off_t slink_totblksize = 0;
slink_base **dupe = NULL;
uint32_t slink_dupe_files = 0;
uint32_t slink_dupe_maxfiles = 0;
uint32_t slink_dupe_zerosizefiles = 0;
off_t slink_dupe_totsize = 0;
off_t slink_dupe_totblksize = 0;
char **slink_dir = NULL;
uint32_t slink_dirs = 0;
uint32_t slink_maxdirs = 0;
uint32_t slink_hardlinks = 0;
uint32_t slink_symlinks = 0;

char **slink_env = NULL;
char rdir[MAX_FILENAME_SIZE];
char rfile[MAX_FILENAME_SIZE];
char repository[MAX_FILENAME_SIZE];
char sizestr [0x100];
char sizestr1[0x100];
char sizestr2[0x100];
uint8_t smd5[16];
char md5_string[33];
char options[] = "acfpqR:rstu";
uint32_t opts = 0;
uint32_t opt_scan = 0;
uint32_t opt_create = 0;
uint32_t opt_packages = 0;
uint32_t opt_restore = 0;
uint32_t opt_repository = 0;
uint32_t opt_quiet = 0;
uint32_t opt_verbose = 0;

char *line = NULL;
size_t line_size = 0;
uint32_t line_no = 1;

void print_message( const char *fmt, ... ) {
	va_list ap;

	va_start(ap, fmt);
	
	if ( !opt_quiet ) {
		vfprintf(stderr, fmt, ap);
	}
	
	va_end(ap);
}

void __attribute__((noreturn)) print_error( const char *fnc, const char *fmt, ... ) {
	va_list args;

	va_start(args, fmt);
	
	if ( !opt_quiet ) {
		vfprintf(stderr, fmt, args);
		if ( fnc && errno) perror(fnc);
		fflush(stderr);
	}
	
	va_end(args);
	
	exit(1);
}

void *check_malloc( size_t size ) {
	void *p;

	if ( !size ) {
		print_error(__FUNCTION__, "NULL allocation request.\n");
	}

	if ( !(p = malloc(size)) ) {
		print_error(__FUNCTION__, "Allocation failed.\n");
	}

	return p;
}

void *check_realloc( void *p, size_t size ) {
	void *ptr;

	if ( !size ) {
		print_error(__FUNCTION__, "NULL reallocation request.\n");
	}

	if ( !(ptr = realloc(p, size)) ) {
		print_error(__FUNCTION__, "Reallocation failed.\n");
	}

	return ptr;
}

void check_asprintf( char **strp, const char *fmt, ... ) {
	va_list args;

	va_start(args, fmt);

	if ( vasprintf(strp, fmt, args) == -1 || !*strp ) {
		print_error(__FUNCTION__, "Allocation failed.\n");
	}

	va_end(args);
}

void add_file( slink_base **base_ptr, char *fullname, uint32_t psize ) {
	struct stat *st = check_malloc(sizeof(struct stat));
	slink_base *bs = *base_ptr;
	
	if ( slink_files >= slink_maxfiles ) {
		slink_maxfiles += ALLOC;
		*base_ptr = bs = check_realloc(bs, slink_maxfiles*sizeof(slink_base));
	}

	if ( lstat(fullname, st) ) {
		print_error("lstat", "Cannot stat \"%s\"\n", fullname);
	}
	
	bs[slink_files].file = fullname;
	bs[slink_files].purename = fullname+psize;
	bs[slink_files].st = st;
	bs[slink_files++].package = NULL;
	
	slink_totsize += st->st_size;
	slink_totblksize += st->st_blocks<<9;
}

void scan_dir( char *path ) {
	size_t psize = strlen(path);
	struct dirent **nl;
	int32_t n;

	n = scandir(path, &nl, NULL, NULL);

	if ( n >= 0 ) {
		int32_t i;

		for ( i=n-1; i; i-- ) {
			if ( nl[i]->d_type == DT_REG && nl[i]->d_name[0] != '.') {
				char *fullname;

				check_asprintf(&fullname, "%s%s", path, nl[i]->d_name);
				add_file(&base, fullname, psize);
			} else {
				if ( nl[i]->d_type == DT_LNK )
					slink_symlinks++;
			}

			if ( nl[i]->d_type != DT_DIR || nl[i]->d_name[0] == '.' ) {
				free( nl[i] );
				nl[i] = NULL;
			}
		}

		for ( i=n-1; i; i-- ) {
			if ( nl[i] ) {
				if ( nl[i]->d_name[0] != '.' ) {
					char *npath;

					check_asprintf(&npath, "%s%s/", path, nl[i]->d_name);
					scan_dir(npath);
					free(npath);
				}
				free(nl[i]);
			}
		}

		free(nl);
	} else {
		print_error("scandir", "");
	}
}

int32_t size_compare( slink_base *f1, slink_base *f2 ) {
/* NO! You don't want to put here "return f1->st->st_size - f2->st->st_size;" ! */
	if ( f1->st->st_size > f2->st->st_size ) return -1;
	if ( f1->st->st_size < f2->st->st_size ) return 1;

	return 0;
}

int32_t package_compare( slink_base *f1, slink_base *f2 ) {
	if ( f1->package && f2->package ) {
		return strcmp( f1->package, f2->package );
	} else {
		if ( !f1->package ) return 1;
	}
	return -1;
}

uint32_t is_same( uint32_t f1, uint32_t f2 ) {
	if ( base[f1].st->st_size != base[f2].st->st_size ) return 0;

	/* don't consider hardlinks as same separate files */
	if ( base[f1].st->st_ino == base[f2].st->st_ino ) {
		slink_hardlinks++;
		return 0;
	}

	/* don't consider files with different permissions as same */
	if ( base[f1].st->st_mode != base[f2].st->st_mode ) {
		return 0;
	}

	/* don't consider files with different ownership as same */
	if ( base[f1].st->st_uid != base[f2].st->st_uid ||
	     base[f1].st->st_gid != base[f2].st->st_gid ) {
		return 0;
	}

	/* zero-sized files are same */
	if ( !base[f1].st->st_size ) return 1;
	
	{
		FILE *ff1, *ff2;
		int32_t r = 1;
		
		if ( !(ff1=fopen(base[f1].file,"rb")) ) {
			print_message("Couldn't open %s for reading to perform comparison, ignoring...\n", base[f1].file);
			return 0;
		}

		if ( !(ff2=fopen(base[f2].file,"rb")) ) {
			print_message("Couldn't open %s for reading to perform comparison, ignoring...\n", base[f2].file);
			return 0;
		}

		{
			size_t size = base[f1].st->st_size;
			size_t rr, sz, r1, r2;
			uint8_t *m1 = check_malloc(R_BUF);
			uint8_t *m2 = check_malloc(R_BUF);
			md5_context ctx;
			
			memset(smd5, 0, 16);
			md5_starts(&ctx);
			
			for ( sz=size; sz; sz-=rr ) {
				rr = sz>R_BUF ? R_BUF : sz;
				r1 = fread(m1, 1, rr, ff1);
				r2 = fread(m2, 1, rr, ff2);
				if ( r1 != rr || r2 != rr ) {
				    	print_message("Error reading %s when comparing file contents, ignoring...\n",
					              r1!=rr?base[f1].file:base[f2].file);
					free(m2);
					free(m1);
					fclose(ff2);
					fclose(ff1);
					
					return 0;
				}
				if ( (r = memcmp(m1, m2, rr)) ) break;
				md5_update(&ctx, m1, r1);
			}
			
			md5_finish(&ctx, smd5);
			
			free(m2);
			free(m1);
		}
		
		fclose( ff2 );
		fclose( ff1 );
		
		if ( !r ) {
			char num[3];
			memset(md5_string, 0, sizeof(md5_string));
			for (r=0; r<16; r++) {
				snprintf(num, 3, "%02x", smd5[r]);
				md5_string[r+r]   = num[0];
				md5_string[r+r+1] = num[1];
			}
			return 1;
		}
	}
	
	return 0;
}

void print_dupe( uint32_t i ) {
	if ( slink_dupe_files >= slink_dupe_maxfiles ) {
		slink_dupe_maxfiles += ALLOC;
		dupe = check_realloc(dupe, slink_maxfiles*sizeof(slink_base));
	}

	dupe[slink_dupe_files++] = &base[i];
	check_asprintf(&base[i].md5, "%s", md5_string);
	
	if ( !opt_packages ) return;

	if ( !vfork() ) {
		execve("./slink_package", (char*[]){"./slink_package", base[i].file, NULL}, slink_env);
	} else {
		wait(NULL);
	}

	{
		FILE *f = fopen("/tmp/slink_pack","rb");
		off_t o;
		char *package = NULL;
		
		if ( !f ) {
			printf("%s\n", base[i].file);
			return;
		}

		fseeko(f, 0, SEEK_END);
		o = ftello(f);
		fseeko(f, 0, SEEK_SET);

		if ( o ) {
			package = check_malloc(o);
			fread(package, 1, o, f);
			package[o-1] = '\0';
		}

		fclose(f);

		base[i].package = package;

		unlink("/tmp/slink_pack");
	}
}

void add_dupe( uint32_t i ) {
	slink_dupe_totsize += base[i].st->st_size;
	slink_dupe_totblksize += base[i].st->st_blocks<<9;
	
	if ( !base[i].st->st_size )
		slink_dupe_zerosizefiles++;
}

void __attribute__((noreturn)) help( char *progpath ) {
	print_message("slink 0.0.1, Copyright (C) 2005 Jindrich Novy (jnovy@redhat.com)\n"
	              "This program is distributed in the hope that it will be useful,\n"
	              "but WITHOUT ANY WARRANTY; See the GNU GPL for more details.\n\n"
	              "usage: %s -[%s] <directory1> <directory2> ...\n", progpath, options);
	exit(0);
}

void compare() {
	uint32_t i;
	
	print_message("Comparing ...\n");
		
	for ( i=0; i<slink_files-1; ) {
		if ( is_same( i, i+1 ) ) {
			print_message("\n");
			print_dupe(i);
			print_dupe(i+1);
			add_dupe(i+1);
			for ( ++i; i<slink_files-1 && is_same(i, i+1); i++ ) {
				print_dupe(i+1);
				add_dupe(i+1);
			}
		} else { 
			print_message(".");
			i++;
		}
	}
}

void add_directory( char ***dir_ptr, char *path ) {
	char **dir = slink_dir;
	char resolved_path[PATH_MAX];
	size_t n;

	if ( slink_dirs >= slink_maxdirs ) {
		slink_maxdirs += R_DIR;
		*dir_ptr = dir = check_realloc(dir, slink_maxdirs*sizeof(char*));
	}
	
	if ( !realpath(path, resolved_path) ) {
		print_error("realpath", "Error resolving path %s.\n", path);
	}
	
	n = strlen(resolved_path);
	
	if ( resolved_path[n-1] == '/' ) {
		dir[slink_dirs] = check_malloc(n+1);
		memcpy(dir[slink_dirs], resolved_path, n+1);
	} else {
		dir[slink_dirs] = check_malloc(n+2);

		memcpy(dir[slink_dirs], resolved_path, n);
		dir[slink_dirs][n]   = '/';
		dir[slink_dirs][n+1] = '\0';
	}

	slink_dirs++;
}

void parse_args( int argc, char **argv ) {
	static struct option slink_opts[] = {
		{"scan", 0, 0, 's'},
		{"create", 0, 0, 'c'},
		{"packages", 0, 0, 'q'},
		{"quiet", 0, 0, 'q'},
		{"repository", 1, 0, 'R'},
		{"restore", 0, 0, 'r'},
		{0, 0, 0, 0}
	};
	int32_t flag, opt_index = 0, badopt = 0;

	for (; (flag = getopt_long(argc, argv, options, slink_opts, &opt_index)) != -1 && !badopt; ) {
		switch( flag ) {
			case 0:
				printf("option %s ", slink_opts[opt_index].name);
				if ( optarg )
					printf("with arg %s", optarg);
				printf("\n");
				break;
			case 's':
				opt_scan = 1;
				opts++;
				break;
			case 'c':
				opt_create = 1;
				opts++;
				break;
			case 'q':
				opt_quiet = 1;
				break;
			case 'r':
				opt_restore = 1;
				opts++;
				break;
			case 'R':
				opt_repository = 1;
				memset(rfile, 0, MAX_FILENAME_SIZE);
				memset(rdir, 0, MAX_FILENAME_SIZE);
				strncpy(repository, optarg, MAX_FILENAME_SIZE-1);
				repository[MAX_FILENAME_SIZE-1] = '\0';
				{
					char *pathsep = strrchr(optarg, '/');
					strncpy(rfile, pathsep+1, MAX_FILENAME_SIZE-1);
					memcpy(rdir, optarg,
					        pathsep-optarg <= MAX_FILENAME_SIZE-1 ?
						pathsep-optarg : MAX_FILENAME_SIZE-1);
				}
				opts++;
				break;
			case 'p':
				opt_packages = 1;
				break;
			case '?':
			default:
				badopt = 1;
				break;
		}
	}

	if ( badopt ) {
		print_error(NULL, "Bad option, exiting.\n");
	}

	for ( ; optind < argc; optind++ ) {
		add_directory( &slink_dir, argv[optind] );
	}
}

void scan_dirs() {
	uint32_t i;
		
	for ( i=0; i<slink_dirs; i++ ) {
		print_message("Scanning \"%s\" ... ", slink_dir[i]);
		scan_dir(slink_dir[i]);
		print_message("\n");
	}
}

void dump_main_file( gzFile r, char *filename, struct stat *st, uint8_t *md5sum ) {
	uint32_t n;

	gzprintf(r,"%s ", filename);
	for (n=0;n<16; n++) {
		gzprintf(r, "%02x", md5sum[n]);
	}
	gzprintf(r, " %lli %llu %llu %u %u %u %u %llu %lu %lli %li %li %li\n",
		 st->st_size, st->st_dev, st->st_ino, st->st_mode, 
		 st->st_nlink,st->st_uid, st->st_gid, st->st_rdev, 
		 st->st_blksize, st->st_blocks,
		 st->st_atime, st->st_mtime, st->st_ctime);
}

void dump_dupe( gzFile r, slink_base *d ) {
	gzprintf(r,"\t%s %lli %llu %llu %u %u %u %u %llu %lu %lli %li %li %li\n", d->file,
		 d->st->st_size, d->st->st_dev, d->st->st_ino, d->st->st_mode,
		 d->st->st_nlink,d->st->st_uid, d->st->st_gid, d->st->st_rdev, 
		 d->st->st_blksize, d->st->st_blocks,
		 d->st->st_atime, d->st->st_mtime, d->st->st_ctime);
}

void eliminate_dupes( gzFile r ) {
	FILE *f;
	uint32_t i;
	uint8_t md5sum[16];
	uint8_t omd5sum[16] = {"\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0"};
	char *temp = NULL;
	uint8_t *m1 = check_malloc(R_BUF);
	md5_context ctx;

	for ( i=0; i<slink_dupe_files; i++ ) {
		if ( !dupe[i]->st->st_size ) continue;

		if ( !(f=fopen(dupe[i]->file,"rb")) ) {
			print_error("fopen", "");
		}
				
		{
			size_t size = dupe[i]->st->st_size;
			size_t rr, sz, r1;

			md5_starts(&ctx);

			for ( sz=size; sz; sz-=rr ) {
				rr = sz>R_BUF ? R_BUF : sz;
				r1 = fread(m1, 1, rr, f);

				if ( r1 != rr ) {
					print_error("fread", "In file %s\n", dupe[i]->file);
				}

				md5_update(&ctx, m1, r1);
			}

			md5_finish(&ctx, md5sum);

			if ( memcmp(md5sum, omd5sum, 16) ) {
				if ( temp ) free(temp);
				
				check_asprintf(&temp, "%s/.%s"TEMP_SUFFIX, rdir, dupe[i]->purename);

				if ( mkstemp(temp) == -1 ) {
					fclose(f);
					gzclose(r);
					print_error("mkstemp", "Error creating temporary file.\n");
				}

				dump_main_file( r, temp, dupe[i]->st, md5sum );

				if ( rename(dupe[i]->file, temp) == -1 ) {
					fclose(f);
					gzclose(r);
					print_error("rename", "");
				}

				if ( symlink(temp, dupe[i]->file) == -1 ) {
					fclose(f);
					gzclose(r);
					print_error("symlink", "");
				}

				memcpy(omd5sum, md5sum, 16);
			} else {
				if ( unlink(dupe[i]->file) ) {
					fclose(f);
					gzclose(r);
					print_error("unlink", "");
				}

				if ( symlink(temp, dupe[i]->file) ) {
					fclose(f);
					gzclose(r);
					print_error("symlink", "");
				}
			}
			dump_dupe( r, dupe[i] );
		}
		
		fclose(f);
	}
	
	if ( temp ) free(temp);
	free(m1);
}

void create_repository() {
	struct stat st;
	gzFile r;

	if ( stat(rdir, &st) ) {
		print_message("Creating new repository in %s ...\n", repository);
		if ( mkdir(rdir, 0755) ) {
			print_error("mkdir", "");
		}
	}

	if ( stat(repository, &st) ) {
		print_message("Creating new %s ...\n", repository);
	}

	if ( !(r=gzopen(repository,"wb+")) ) {
		print_error("fopen", "");
	}

	eliminate_dupes(r);

	gzclose(r);
}

uint32_t get_line( gzFile f ) {
	if ( !line ) {
		line = check_malloc(MAX_LINE_SIZE+1);
	}
	
	for ( line_size=0; !gzeof(f) && line_size < MAX_LINE_SIZE; line_size++ ) {
		line[line_size] = gzgetc(f);
		if ( line[line_size] == '\n' )
			break;
	}
	
	line[line_size] = '\0';
	
	if ( gzeof(f) ) return 0;
	
	return 1;
}

char *get_string( char *dest, char *src, uint32_t len ) {
	uint32_t i;

	for ( i = 0; src[i] && len-1 > i && !isspace(src[i]); dest[i] = src[i], i++ );
	dest[i] = '\0';

	for ( ; len-1 > i && isspace(src[i]); i++ );
	
	if ( len-1 <= i )
		print_message("String in repository is too long so was truncated.\n");

	return &src[i];
}

char *get_md5( char *dest, char *src ) {
	uint32_t i;

	for ( i = 0; i<32 && isxdigit(src[i]); dest[i] = src[i], i++ );
	dest[i] = '\0';

	if ( i != 32 ) {
		print_error("get_md5", "Invalid MD5 sum in the repository -> corrupted.\n");
	}

	for ( ; isspace(src[i]); i++ );

	return &src[i];
}

void restore_from_repository( char *rep_dir, char *rep_file ) {
	char *rep;
	gzFile r;

	check_asprintf(&rep, "%s/%s", rep_dir, rep_file);

	if ( !(r=gzopen(rep, "rb")) ) {
		print_error("fopen", "Error opening repository %s\nHave you created one already?\n", rep);
	}

	while ( get_line(r) ) {
		if ( line_size ) {
			char ori_file[MAX_LINE_SIZE], linked[MAX_LINE_SIZE];
			char md5sum[33];
			char *p = line;
			struct stat st;
			uint32_t d;
			z_off_t o;

			p = get_string(ori_file, p, MAX_LINE_SIZE);
			p = get_md5(md5sum, p);

			line_no++;

			for ( d=0; !gzeof(r); d++ ) {
				o = gztell(r);
				get_line(r);

				if ( line[0] != '\t' ) {
					if ( d<2 ) {
						print_message("Don't handcraft the repository! Only one or no link found.\n");
					}
					gzseek(r, o, SEEK_SET);
					break;
				}

				p = get_string( linked, line+1, MAX_LINE_SIZE );
				if ( sscanf(p, "%lli %llu %llu %u %u %u %u %llu %lu %lli %li %li %li",
					&st.st_size, &st.st_dev, &st.st_ino, &st.st_mode,
					&st.st_nlink,&st.st_uid, &st.st_gid, &st.st_rdev,
					&st.st_blksize, &st.st_blocks,
					&st.st_atime, &st.st_mtime, &st.st_ctime) != 13 ) {
					print_error(NULL, "Error while scanning stat info.\n");
				}

				line_no++;

				if ( opt_verbose ) print_message("Restoring %s ...\n", linked);

				{
					struct stat stt;

					if ( !lstat(linked, &stt) ) {
						if ( !S_ISLNK(stt.st_mode) ) {
							print_error("lstat", "Repository record refers to non-symlink!\n");
						}

						if ( unlink(linked) ) {
							print_error("unlink", "");
						}
					} else {
						print_message("lstat: Cannot stat linked file %s.\n", linked);
					}
				}

				{
					FILE *f1, *f2;
					size_t rr, sz, r1, w2;
					uint8_t *m1 = check_malloc(R_BUF);

					if ( !(f1=fopen(ori_file,"rb")) ) {
						print_error("fopen", "Error opening original file in repository!\n");
					}

					if ( !(f2=fopen(linked,"wb")) ) {
						print_error("fopen", "Error opening reconstructed file for writing!\n");
					}

					for ( sz=st.st_size; sz; sz-=rr ) {
						rr = sz>R_BUF ? R_BUF : sz;
						r1 = fread(m1, 1, rr, f1);

						if ( r1 != rr ) {
							print_error("fread", "in file %s\n", ori_file);
						}

						w2 = fwrite(m1, 1, rr, f2);

						if ( w2 != rr ) {
							print_error("fwrite", "in file %s\n", linked);
						}
					}
					
					free(m1);

					fclose(f2);
					fclose(f1);

					if ( chmod(linked, st.st_mode) ) {
						print_message("Failed to set original permissions for %s.\n", linked);
					}

					if ( chown(linked, st.st_uid, st.st_gid) ) {
						print_message("Failed to set original uid/gid for %s.\n", linked);
					}

					if ( utime(linked, &(struct utimbuf){st.st_atime, st.st_mtime}) ) {
						print_message("Failed to set original actime/modtime for %s.\n", linked);
					}
				}
			}
		}
	}

	gzclose(r);

	if ( line ) free(line);

	free(rep);
}

void snprint_size( char *s, size_t ss, off_t size ) {
	if ( size<10000 ) {
		snprintf(s, ss, "%4llu   B", size);
	} else {
		if ( size<10000<<10 ) {
			snprintf(s, ss, "%4llu KiB", size>>10);
		} else {
			snprintf(s, ss, "%4llu MiB", size>>20);
		}
	}
}

void snprint_size_unaligned( char *s, size_t ss, off_t size ) {
	if ( size<10000 ) {
		snprintf(s, ss, "%llu B", size);
	} else {
		if ( size<10000<<10 ) {
			snprintf(s, ss, "%llu KiB", size>>10);
		} else {
			snprintf(s, ss, "%llu MiB", size>>20);
		}
	}
}

char *print_size_duplet( off_t s, off_t sb, int style ) {
	if ( style == SLINK_ALIGNED ) {
		snprint_size(sizestr1, sizeof(sizestr1), s);
		snprint_size(sizestr2, sizeof(sizestr2), sb);
	} else {
		snprint_size_unaligned(sizestr1, sizeof(sizestr1), s);
		snprint_size_unaligned(sizestr2, sizeof(sizestr2), sb);
	}
	
	snprintf(sizestr, sizeof(sizestr), "%s [%s]", sizestr1, sizestr2);
	
	return sizestr;
}

void print_results() {
	fflush(stdout);
	printf("\n%u regular files found, %s\n",
	       slink_files, print_size_duplet(slink_totsize, slink_totblksize, SLINK_UNALIGNED));
	printf("%u total dupes, %u non-zero sized.\n", 
	       slink_dupe_files, slink_dupe_files-slink_dupe_zerosizefiles);
	printf("%s, %.3f%% [%.3f%%] wasted by dupes, %u symlinks, %u hardlinks.\n",
	       print_size_duplet(slink_dupe_totsize, slink_dupe_totblksize, SLINK_UNALIGNED),
	       slink_totsize ? slink_dupe_totsize*100./slink_totsize : 0.,
	       slink_totblksize ? slink_dupe_totblksize*100./slink_totblksize: 0.,
	       slink_symlinks, slink_hardlinks);
}

int main( int argc, char **argv, char **envp ) {
	slink_env = envp;

	parse_args( argc, argv );

	if ( !opts ) {
		if ( slink_dirs ) {
			opts++;
			opt_scan=1;
		} else {
			help(argv[0]);
		}
	}

	if ( opts!=1 ) {
		print_message("You specified some conflicting options (some options twice?).\n");
		exit(1);
	}
	
	if ( opt_scan || opt_create ) {
		scan_dirs();

		if ( !slink_files ) {
			print_message("No regular files found.\n");
			exit(1);
		}

		qsort(base, slink_files, sizeof(slink_base), (int32_t (*)(const void*,const void*))size_compare);
		
		compare();
	}

	if ( !opt_repository ) {
		if ( !getuid() ) {
			strcpy(repository, "/etc/slink/.repository.gz");
			strcpy(rdir, "/etc/slink");
		} else {
			strcpy(repository, "~/.slink/.repository.gz");
			strcpy(rdir, "~/.slink");
		}
		strcpy(rfile, ".repository.gz");
	}

	if ( opt_create ) {
		if ( !slink_dupe_files ) {
			print_message("No dupes, repository creation aborted.\n");
			exit(0);
		}

		create_repository();
	}
	
	if ( opt_restore ) {
		restore_from_repository(rdir, rfile);
		return 0;
	}
	
	if ( opt_packages ) {
		uint32_t i, p = 0, f = 0;
		off_t s = 0, bs = 0;
		size_t len = 0;
		
		printf("\n");
		
		qsort(base, slink_files, sizeof(slink_base), (int32_t (*)(const void*,const void*))package_compare);
		
		for ( i=0; i<slink_files; i++ )
			if ( base[i].md5 && strlen(base[i].file) > len )
				len = strlen(base[i].file);

		if ( slink_files && base[0].package ) {
			printf("%s:\n", base[0].package);
			printf("%s%*s\t%s\t%s\n",
			        base[0].file, len - strlen(base[0].file), "",
				print_size_duplet(base[0].st->st_size, base[0].st->st_blocks<<9, SLINK_ALIGNED),
				base[0].md5);
			s  = base[0].st->st_size;
			bs = base[0].st->st_blocks<<9;
			p++;
		}
		
		for ( i=1; base[i].package && i<slink_files; i++ ) {
			if ( strcmp(base[i-1].package, base[i].package) ) {
				printf("Disc space wasted: %s\n\n", print_size_duplet(s, bs, SLINK_UNALIGNED));
				printf("%s:\n", base[i].package);
				printf("%s%*s\t%s\t%s\n",
			        	base[i].file, len - strlen(base[i].file), "",
					print_size_duplet(base[i].st->st_size, base[i].st->st_blocks<<9, SLINK_ALIGNED),
					base[i].md5);
				s  = base[i].st->st_size;
				bs = base[i].st->st_blocks<<9;
				p++;
			} else {
				printf("%s%*s\t%s\t%s\n",
			        	base[i].file, len - strlen(base[i].file), "",
					print_size_duplet(base[i].st->st_size, base[i].st->st_blocks<<9, SLINK_ALIGNED),
					base[i].md5);
				s  += base[i].st->st_size;
				bs += base[i].st->st_blocks<<9;
			}
		}
		printf("Disc space wasted: %s\n\n", print_size_duplet(s, bs, SLINK_UNALIGNED));
		
		printf("Total packages with duplicates: %d\n", p);

		for ( s=bs=0; i<slink_files; i++ ) {
			if ( !base[i].md5 ) continue;

			if ( !f ) {
				printf("\nDupes with no package:\n");
				f = 1;
			}
			
			printf("%s%*s\t%s\t%s\n",
			       	base[i].file, len - strlen(base[i].file), "",
				print_size_duplet(base[i].st->st_size, base[i].st->st_blocks<<9, SLINK_ALIGNED),
				base[i].md5);
			s  += base[i].st->st_size;
			bs += base[i].st->st_blocks<<9;
		}
		
		if ( f ) printf("Disc space wasted: %s\n", print_size_duplet(s, bs, SLINK_UNALIGNED));
	} else {
		uint32_t i;
		size_t len = 0;

		printf("\n\n");

		for ( i=0; i<slink_files; i++ )
			if ( base[i].md5 && strlen(base[i].file) > len )
				len = strlen(base[i].file);

		for ( i=0; i<slink_files; i++ ) {
			if ( !base[i].md5 ) continue;

			printf("%s%*s\t%s\t%s\n",
			       	base[i].file, len - strlen(base[i].file), "",
				print_size_duplet(base[i].st->st_size, base[i].st->st_blocks<<9, SLINK_ALIGNED),
				base[i].md5);
			fflush(stdout);
		}
	}
	
	print_results();
	
	return 0;
}
